/*
    Filename: main.c
    Synopsis: Main module of PA Window Manager.

                          Copyright (C) 2000-2006
            Ral Nez de Arenas Coronado <raul@pleyades.net>
	         David Gmez Espinosa <david@pleyades.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA 02111-1307 USA

*/

#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "config.h"
#include "x.h"
#include "wm.h"
#include "mobs.h"
#include "signals.h"
#include "palabel.h"


/* Global object for the config pool */
CfgPool pawmpool;
#ifdef ENABLE_STARTUP
extern bool volatile startup_timeout;
#endif

static void term_handler ();
static void chld_handler ();



/* For terminating politely */
static void term_handler () 
{
    /* Clean all the window manager mess ;) */
    wm_done();

    x_done();

    SAY("Deleting Pawm Config pool");
    cfgpool_delete(pawmpool);

    /* Farewell, thy the puto amo of all window managers */
    exit(EXIT_SUCCESS);
}


#ifdef ENABLE_STARTUP
static void alrm_handler () 
{
	SAY("Startup timeout expiring");
	
	startup_timeout=true;
}
#endif


/* Wait for our beloved childrens ;) (X clients launched by pimp) */
static void chld_handler () 
{
    waitpid(-1,NULL,WNOHANG);
}


/* Get config from ~/.pawm or /etc */
static void get_pawm_conf() 
{
	char *home; 
	char *tmpname=NULL;

	pawmpool=NULL;

	home=getenv("HOME");
	if (home!=NULL) {
		tmpname=malloc(strlen(home)+strlen("/.pawm/pawm.conf")+1);
		strcpy(tmpname, home); strcat(tmpname, "/.pawm/pawm.conf");
		if (access(tmpname, F_OK)==0) {
			pawmpool=cfgpool_create(tmpname);
		}
	}
	if (pawmpool==NULL) {
		if (tmpname!=NULL)
			free(tmpname);
		tmpname=malloc(strlen(CONFDIR)+strlen("/pawm.conf")+1);
		strcpy(tmpname, CONFDIR); strcat(tmpname, "/pawm.conf");
		pawmpool=cfgpool_create(tmpname);
	}
	if (pawmpool==NULL) {
		if (errno==ENOENT) die("%s doesn't exist",tmpname);
		else BUG();
	}
	
	free(tmpname);
}


int main (int argc, char *argv[]) 
{
	struct sigaction sa_term, sa_chld, sa_alrm;
    
    /* Command line parameters */
	if (argc > 1) {    
		if ( strcmp(argv[1],"--version")==0 ) {
			printf("%s-%s\n", PROJECT, VERSION);
			printf("Report bugs to: David Gmez <david@pleyades.net>\n");
			exit(EXIT_SUCCESS);
		} else if ( strcmp(argv[1],"--display")==0 ) {
			if (argv[2]) {
				display_name=malloc(strlen(argv[2]+1));
				strcpy(display_name,argv[2]);
			} else {
				die("Parameter unknown. %s [--display dpy_name | --version]", progname(argv[0]));
			}
		}
	}

	block_signals();

	get_pawm_conf();

	SAY("Starting %s %s", PROJECT, VERSION);

	x_init();

	/* Just for killing the window manager. X also get killed if there are
	 * not X wrappers */
	sigemptyset(&sa_term.sa_mask);
	sa_term.sa_flags=0;
	sa_term.sa_handler=term_handler;
	if ( sigaction(SIGTERM, &sa_term, (struct sigaction *) 0)==-1 ) BUG();
	if ( sigaction(SIGINT, &sa_term, (struct sigaction *) 0)==-1 ) BUG();

	sigemptyset(&sa_chld.sa_mask);
	sa_chld.sa_flags=0;
	sa_chld.sa_handler=chld_handler;
	if ( sigaction(SIGCHLD, &sa_chld, (struct sigaction *) 0)==-1 ) BUG();
	
#ifdef ENABLE_STARTUP
	sigemptyset(&sa_alrm.sa_mask);
	sa_alrm.sa_flags=0;
	sa_alrm.sa_handler=alrm_handler;
	if ( sigaction(SIGALRM, &sa_alrm, (struct sigaction *) 0)==-1 ) BUG();
#endif

	if (!wm_init()) {
		x_done();
		return 0;
	}

	/* Get ya motor running, headin' on tha highway... */
	x_main_loop();
	
	/* The last loop NEVER returns... */
	return 0;
}
