/*
  Filename: pimp.c  
  Synopsys: An Application Launcher
    
 	              Copyright (C) 2001-2006
	         David Gmez Espinosa <david@pleyades.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/time.h>
#include <sys/stat.h>
#include "pimp.h"
#include "paicon.h"


#define BUFSIZE 256


static struct {
    char *confdir;
} class;


static void pimp_findapps(void);

/* Global object for keeping apps info */
static Chain ThePimp;



void pimp_init()
{
    struct stat sinfo;
    char *homedir;
    
    homedir=getenv("HOME");    
    class.confdir=malloc(strlen(homedir)+strlen(".pawm")+1);
    if (!class.confdir) die("Not enough memory");

    ThePimp=chain_create();
    if (!ThePimp) BUG();

    paicon_init();
    
    sprintf(class.confdir,"%s/.pawm",homedir);
    
    /* Check if pawm directory exists */
    if ( (stat(class.confdir,&sinfo))==-1 ) {
	bang("stat failed");
	return;
    }
    if (!S_ISDIR(sinfo.st_mode)) {
	bang("%s in not a directory", class.confdir);
	return;
    }

    pimp_findapps();
    pimp_mapall();
}



void pimp_done()
{
	PAIcon somehooker;
	int x,y;
	char *file;
	FILE *fp;
	char tmpbuf[BUFSIZE];
	char line1[BUFSIZE], line2[BUFSIZE];
	char line3[BUFSIZE], line4[BUFSIZE];

	chain_seek(ThePimp, chain_size(ThePimp));
	while ( (somehooker=chain_del(ThePimp))!=NULL ) {

		/* Save the icon position back to the icon config file */
		paicon_getfile(somehooker, &file);
		sprintf(tmpbuf,"%s/%s",class.confdir, file);

		if ( !(fp=fopen(tmpbuf,"r")) ) {
			bang("File %s cannot be opened\n",tmpbuf);
			continue;
		}

		/* Get the current data in the file... */
		fgets(line1, BUFSIZE, fp);
		fgets(line2, BUFSIZE, fp);
		fgets(line3, BUFSIZE, fp);
		fgets(line4, BUFSIZE, fp);

		/* truncate the file... */
		fclose(fp);
		fp=fopen(tmpbuf,"w");

		/* get the position... */
		paicon_getpos(somehooker,&x,&y);

		/* ... and write it all, with the new position */
		fputs(line1, fp);
		fprintf(fp,"%d %d\n",x,y);
		fputs(line3, fp);
		fputs(line4, fp);

		fflush(fp); 
		fclose(fp);

		/* And finally free the icon object */
		paicon_destroy(somehooker);
	}

	chain_delete(ThePimp);

	paicon_done();
	free(class.confdir);

}


/* This function should be called only one time, when the window manager is
 * initialized, and icons will stay mapped until X are closed */
void pimp_mapall()
{
    PAIcon someicon;	
    
    chain_home(ThePimp);
    while ( (someicon=chain_next(ThePimp))!=NULL ) {
	/* We assign a handler for the hooker (hey, that sounds weird ;) )
	 * and map the button */
	paicon_mapall(someicon);
    }

}


/* Scan the config dir, looking for iconf config files, if any, and create
   the appropiate icons that are stored in a Chain */
static void pimp_findapps()
{
    DIR *pawmdir;               /* Per-user config dir */   
    struct dirent *dirdata;     /* Used for icons config files */
    FILE *fp;
    char *tmpbuf;	        /* Temp buffer for data */	
    PAIcon meretriz;            /* Used to read the apps that 
			           are stored in the chain */
    xImage image;
    unsigned int tmpx, tmpy;
 
    tmpbuf=malloc(BUFSIZE);
    if (!tmpbuf) die("Not enough memory");

    /* If we're here, directory exists, so if opendir fails miserably,
       we'll die miserably :)) */
    if ( (pawmdir=opendir(class.confdir))==NULL )  BUG(); 

    while ( (dirdata=readdir(pawmdir))!=NULL ) {
	    
        if (!strcmp(dirdata->d_name,".")) continue;
        if (!strcmp(dirdata->d_name,"..")) continue;
	if (strncmp(dirdata->d_name,"app",3)!=0) continue;
	if (dirdata->d_name[strlen(dirdata->d_name)-1]=='~') continue;
	
	/* Build path */
        snprintf(tmpbuf,strlen(class.confdir)+strlen(dirdata->d_name)+2,
   	         "%s/%s",class.confdir,dirdata->d_name);
        SAY("Opening %s",tmpbuf);

	/* And look in the guts of each file */
	if ( !(fp=fopen(tmpbuf,"r")) ) BUG();

	/* Get the icon name */
	if (fgets(tmpbuf,BUFSIZE,fp)==NULL) {
	   bang("Failure reading app file %s",dirdata->d_name);
	   continue;
	}
	   
	tmpbuf[strlen(tmpbuf)-1]='\0';
	
    image=x_build_pixmap(tmpbuf);
	if (image==NULL) {
	    bang("Icon %s does not exists", tmpbuf);
		image=x_build_pixmap(DEFAULT_ICON);
		if (image==NULL) die("Default icon not found!");
	}
	meretriz=paicon_create(image);
	paicon_setfile(meretriz, dirdata->d_name);
       
	/* Get the position */
	if (fgets(tmpbuf,BUFSIZE,fp)==NULL) {
		bang("Failure reading app file %s",dirdata->d_name);
		paicon_destroy(meretriz);
		continue;
	}
	if (sscanf(tmpbuf,"%d %d",&tmpx,&tmpy)==0) {
	    bang("Bad format in app file %s",dirdata->d_name);
	    paicon_destroy(meretriz);
	    continue;
	}
	paicon_seticonpos(meretriz,tmpx,tmpy);

	/* Get the icon text */
	if (fgets(tmpbuf,BUFSIZE,fp)==NULL) {
	    bang("Failure reading app file %s",dirdata->d_name);
	    paicon_destroy(meretriz);
	    continue;
	}
	tmpbuf[strlen(tmpbuf)-1]='\0';
	paicon_seticontext(meretriz,tmpbuf);

	/* Get the name for the app binary */
	if (fgets(tmpbuf,BUFSIZE,fp)==NULL) {
	    bang("Failure reading app file %s",dirdata->d_name);
	    paicon_destroy(meretriz);
	    continue;
	}

	tmpbuf[strlen(tmpbuf)-1]='\0';
	paicon_setapp(meretriz, tmpbuf);
	
#ifdef ENABLE_STARTUP
	/* Get startup notify field */
	if (fgets(tmpbuf,BUFSIZE,fp)!=NULL) {
		tmpbuf[strlen(tmpbuf)-1]='\0';
		if (strcmp(tmpbuf,"false")==0) {
			paicon_disable_startup(meretriz);
		}	
	}
#endif

	
	/* Add new hooker to chain */
	chain_put(ThePimp,meretriz);

	fclose(fp);
    }
    
	closedir(pawmdir);
	free(tmpbuf);
}

