/*
  Filename: signals.c  
  Synopsys: Some misc signal handling functions
    
 	                      Copyright (C) 2001-2006  
            Ral Nuez de Arenas Coronado <raul@pleyades.net>
	         David Gmez Espinosa <david@pleyades.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/types.h>
#include "signals.h"

/* Non-ignored signals. This array marks its end with '0',
 * since this value cannot be a legal signal number. */
int non_ignored_sigs[]={
    SIGSEGV,
    SIGTRAP,
    SIGABRT,
    SIGILL,
    SIGBUS,
    SIGFPE,
    SIGTERM,
    SIGCONT,
    SIGSTOP,    
    SIGALRM,
    SIGHUP,
    SIGCHLD,
    SIGINT,
    0
};


void block_signals()
{
	size_t i=0;
	sigset_t bsignals;

	sigfillset(&bsignals);
	/* Don't block this signals */
	while(non_ignored_sigs[i++] != 0 )
		sigdelset(&bsignals,non_ignored_sigs[i]);
	sigprocmask(SIG_BLOCK,&bsignals,0);
}


void unblock_signals()
{
	size_t i=0;
	sigset_t bsignals;

	sigfillset(&bsignals);
	while(non_ignored_sigs[i++] != 0 )
		sigdelset(&bsignals,non_ignored_sigs[i]);
	sigprocmask(SIG_UNBLOCK,&bsignals,0);
}

