/*
    Filename: x.c
    Synopsis: X framework.

                          Copyright (C) 2000-2006
            Ral Nez de Arenas Coronado <raul@pleyades.net>
	         David Gmez Espinosa <david@pleyades.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA 02111-1307 USA

*/

#include "config.h"
#include "x.h"

#include <sys/select.h>

#define BUFSIZE 1024
#define PIXMAPDIR "/usr/share/pixmaps"
#define MIN_PIXMAPW 16
#define MIN_PIXMAPH 16
#define MAX_PIXMAPW 64
#define MAX_PIXMAPH 64

#define ICONW 48
#define ICONH 48

/* Global resources commonly used */
GC xgc=None;
Display *display=NULL;
int displayfd=-1;
#ifdef ENABLE_STARTUP
SnDisplay *sn_display=NULL;
SnMonitorContext *sn_context=NULL;
volatile bool startup_timeout=false;
static short int seq_ref_count=0;  /* Reference counter for startup-sequences */
#endif
char *display_name=NULL;
Cursor x_root_cursor=None;
Cursor clockcursor=None;
Error error={
	code:0,
	msg:NULL,
	natural:0,
	integer:0,
	real:0,
	pointer:NULL,
};

/* This struct defines a widget callback.  These callbacks are called whenever
 * an event arrives whose destination is the window in the Sloot, and is called
 * using the 'widget' as parameter. */
typedef struct {
	Window window;
	Handler handle;
	void *widget;
} Sloot;

static Chain sloots;  /* Stores the Sloots */

/* Startup notification handler */
#ifdef ENABLE_STARTUP
static void snevent_handler(SnMonitorEvent *, void *);
#endif
/* Our handlers for X errors */
static int _errorhandler   (Display *, XErrorEvent *);
static int _ioerrorhandler (Display *);

static unsigned short int reference_count=0;





/* This function initializes our fantastic X framework ;) */
void x_init (void) 
{
    XVisualInfo vinfo;

    if (reference_count++) return;

    display = XOpenDisplay(NULL);
    if (!display) die("Pawm is unable to open display");

	displayfd = ConnectionNumber(display);

#ifdef ENABLE_STARTUP
	sn_display=sn_display_new(display, NULL, NULL);
	sn_context = sn_monitor_context_new (sn_display, SCREENI, snevent_handler, NULL, NULL);
#endif	
    
    /* Deactivate screen saver */
    XSetScreenSaver(display, 0, 0, DefaultBlanking, DefaultExposures);

    XStoreName(display, ROOT, "root");

    /* The default GC */        
    xgc=XCreateGC(display, ROOT, 0, NULL);
    if (xgc==None) BUG();

    /* Setup colors */
    if (!XMatchVisualInfo(display, SCREENI, DEPTH, TrueColor, &vinfo)) BUG();

    x_root_cursor=XCreateFontCursor(display, XC_left_ptr);
	clockcursor=XCreateFontCursor(display, XC_watch);
    
    XDefineCursor(display, ROOT, x_root_cursor);

    /* Install the error handlers */
    XSetErrorHandler(_errorhandler);
    XSetIOErrorHandler(_ioerrorhandler);

    /* Create the Sloot list */
    sloots=chain_create();
    if (!sloots) BUG();

	/* ICCCM atoms */
	atoms.string=XInternAtom(display, "STRING", False);
	atoms.wm_name=XInternAtom(display, "WM_NAME", False);
	atoms.wm_icon_name=XInternAtom(display, "WM_ICON_NAME", False);
	atoms.wm_normal_hints=XInternAtom(display, "WM_NORMAL_HINTS", False);
	atoms.wm_transient_for=XInternAtom(display, "WM_TRANSIENT_FOR", False);
	atoms.wm_state=XInternAtom(display, "WM_STATE", False);
	atoms.wm_delete_window=XInternAtom(display, "WM_DELETE_WINDOW", False);
	atoms.wm_protocols=XInternAtom(display, "WM_PROTOCOLS", False);
	atoms.wm_change_state=XInternAtom(display, "WM_CHANGE_STATE", False);
	/* EWMH */
	atoms.utf8_string=XInternAtom(display, "UTF8_STRING", False);
	atoms.net_wm_name=XInternAtom(display, "_NET_WM_NAME", False);
	atoms.net_wm_icon_name=XInternAtom(display, "_NET_WM_ICON_NAME", False);
	atoms.net_wm_window_type=XInternAtom(display, "_NET_WM_WINDOW_TYPE", False);
	atoms.net_wm_window_type_desktop=XInternAtom(display, "_NET_WM_WINDOW_TYPE_DESKTOP", False);
	atoms.net_wm_window_type_dock=XInternAtom(display, "_NET_WM_WINDOW_TYPE_DOCK", False);
	atoms.net_wm_window_type_toolbar=XInternAtom(display, "_NET_WM_WINDOW_TYPE_TOOLBAR", False);
	atoms.net_wm_window_type_menu=XInternAtom(display, "_NET_WM_WINDOW_TYPE_MENU", False);
	atoms.net_wm_window_type_utility=XInternAtom(display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
	atoms.net_wm_window_type_splash=XInternAtom(display, "_NET_WM_WINDOW_TYPE_SPLASH", False);
	atoms.net_wm_window_type_dialog=XInternAtom(display, "_NET_WM_WINDOW_TYPE_DIALOG", False);
	atoms.net_wm_window_type_normal=XInternAtom(display, "_NET_WM_WINDOW_TYPE_NORMAL", False);
	atoms.net_wm_state=XInternAtom(display, "_NET_WM_STATE", False);
	atoms.net_wm_state_modal=XInternAtom(display, "_NET_WM_STATE_MODAL", False);
	atoms.net_wm_state_sticky=XInternAtom(display, "_NET_WM_STATE_STICKY", False);
	atoms.net_wm_state_maximized=XInternAtom(display, "_NET_WM_STATE_MAXIMIZED", False);
	atoms.net_wm_state_shaded=XInternAtom(display, "_NET_WM_STATE_SHADED", False);
	atoms.net_wm_state_skip_taskbar=XInternAtom(display, "_NET_WM_STATE_SKIP_TASKBAR", False);
	atoms.net_wm_state_skip_pager=XInternAtom(display, "_NET_WM_SKIP_PAGER", False);
	atoms.net_wm_state_hidden=XInternAtom(display, "_NET_WM_STATE_HIDDEN", False);
	atoms.net_wm_state_fullscreen=XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", False);
	atoms.net_wm_state_above=XInternAtom(display, "_NET_WM_STATE_ABOVE", False);
	atoms.net_wm_state_below=XInternAtom(display, "_NET_WM_STATE_BELOW", False);
	atoms.net_supported=XInternAtom(display, "_NET_SUPPORTED", False);
	atoms.net_current_desktop=XInternAtom(display, "_NET_CURRENT_DESKTOP", False);
	atoms.net_active_window=XInternAtom(display, "_NET_ACTIVE_WINDOW", False);
	atoms.net_workarea=XInternAtom(display, "_NET_WORKAREA", False);
	atoms.net_supporting_wm_check=XInternAtom(display, "_NET_SUPPORTING_WM_CHECK", False);
	atoms.net_number_of_desktops=XInternAtom(display, "_NET_NUMBER_OF_DESKTOPS", False);
	atoms.net_client_list=XInternAtom(display, "_NET_CLIENT_LIST", False);
	atoms.net_wm_icon=XInternAtom(display, "_NET_WM_ICON", False);
	/* Other atoms */
	atoms.root_pixmap=XInternAtom(display, "_XROOTPMAP_ID", False);
	atoms.motif_wm_hints=XInternAtom(display,"_MOTIF_WM_HINTS", False);

}




/* This function shuts down our fantastic X framework */
void x_done (void) 
{
	Sloot *victim=NULL;

	if (--reference_count) return;

	/* Let our sloots run away free ;), they did a good work */
	chain_seek(sloots, chain_size(sloots));
	while (chain_size(sloots)) {
		victim=chain_del(sloots);
		free(victim);
	}

#ifdef ENABLE_STARTUP
	sn_monitor_context_unref(sn_context);
	sn_display_unref(sn_display);
#endif

	/* Destroy the chain itself */
	chain_delete(sloots);

	/* Free resources... */
	XFreeGC(display, xgc);
	XFreeCursor(display, x_root_cursor);

	XCloseDisplay(display);
}




/* This function returns a pixel value from a color name which value
 * is specified in the pawm config file */
unsigned long x_color(const char *color) 
{
	XColor scolor,ecolor;
	char *tmpvalue;

	tmpvalue=cfgpool_getstring(pawmpool,color);
	/* White as default if not color is found */
	if (tmpvalue==NULL) { 
		SAY("Color %s not found in config file",color);
		return WHITE;
	}
	XAllocNamedColor(display,COLORMAP,tmpvalue,&scolor,&ecolor);
	free(tmpvalue);

	return scolor.pixel;
}


char *x_colorstr(const char *color) 
{
	char *tmpvalue;

	tmpvalue=cfgpool_getstring(pawmpool,color);
	/* White as default if not color is found */
	if (tmpvalue==NULL) { 
		SAY("Color %s not found in config file",color);
		return "white";
	}

	return tmpvalue;
}


/* This function returns a pixel value from an RGB specification */
unsigned long x_pixel(const char *pixel) 
{
	XColor scolor,ecolor;

	XAllocNamedColor(display,COLORMAP,pixel, &scolor,&ecolor);

	return scolor.pixel;
}


/* This function creates a pixmap from a icon in xpm format, also returning
 * an icon mask for XPM's with transparency. All the data is contained in
 * a xImage type */
xImage x_build_pixmap(char *xpmicon)
{
	XpmAttributes attr;
	char xpmpath[BUFSIZE];
	struct stat sinfo;
	int rvalue;
	xImage image;
	char *end_xpmpath;

	ASSERT(xpmicon);

	snprintf(xpmpath, BUFSIZE, "%s/icons/%s", DATADIR, xpmicon);
	if ( stat(xpmpath, &sinfo)==-1 ) {
		snprintf(xpmpath, BUFSIZE, "%s/%s", PIXMAPDIR, xpmicon);
		if (stat(xpmpath, &sinfo)==-1) return NULL;
	}

	image=malloc(sizeof(struct sxImage));
	image->pixmap=None; image->mask=None;

	/* Use default values */
	attr.valuemask=0;
	rvalue=XpmReadPixmapFile(display,ROOT,xpmpath, &(image->pixmap),&(image->mask),&attr);    
	if (rvalue!=XpmSuccess) {
		/* Fallback to 256 color icons...*/
		SAY("Using 256 color icons...");

		end_xpmpath=xpmpath+strlen(xpmpath)-4;
		strcat(end_xpmpath,"_256.xpm");
		rvalue=XpmReadPixmapFile(display,ROOT,xpmpath, &(image->pixmap),&(image->mask),&attr);    
		if (rvalue!=XpmSuccess) {
			bang("XPM cannot be created");
			free(image);
			return NULL;
		}    
	}

	/* Check size limits for XPM files. They're reasonable limits for
	   all resolutions, although they can be changed without problems */
	if ( (attr.width < MIN_PIXMAPW) || (attr.width > MAX_PIXMAPW) ||
		 (attr.height < MIN_PIXMAPH) || (attr.height > MAX_PIXMAPH)) {
		fprintf(stderr, "%s xpm size is too big or too small", xpmpath);
	}

	image->w=attr.width; image->h=attr.height;

	XpmFreeAttributes(&attr);

	return image;
}


xImage x_build_pixmap_from_argb(void *data, size_t len)
{
	XImage *image=NULL;
	xImage icon=NULL;
	CARD32 **icons;
	CARD32 offset=0, idx=0;
	unsigned char *buffer_img; int img_idx;
	unsigned long frame_color;
	int i;
	float alpha, r, g, b;
	unsigned int rb, gb, bb;
	
	ASSERT(icons);

	frame_color=x_color("frame_base_color");
	
	icons= (CARD32 **)data;
	
    icon=malloc(sizeof(struct sxImage));
    icon->pixmap=None; icon->mask=None;;

	do {
		offset=idx;
		icon->w=(int)icons[idx];
		icon->h=(int)icons[idx+1];
		SAY("Icon with size %d %d", icon->w, icon->h);
		
		/* We found the desired size , if not get the next one */
		if ((icon->w==ICONW) && (icon->h==ICONH)) break;
		
		idx+=icon->w*icon->h+2;
	 } while(idx < len); 
	
	buffer_img=malloc(icon->w*icon->h * sizeof(CARD32));
	img_idx=0;
	
	/* Advance to the start of the icon data */
	icons+=offset+2;

	rb= ((frame_color) & 0xFF0000) >> 16; 
	gb= ((frame_color) & 0x00FF00) >> 8; 
	bb= ((frame_color) & 0x0000FF); 
	
	for (i=0; i < icon->w*icon->h; i++) {
		alpha= ((((CARD32)icons[0]) >> 24))/255.0; 
		/* Pre-multiplied alpha */
		r= ((((CARD32)icons[0]) & 0x00FF0000) >> 16)*alpha; 
		g= ((((CARD32)icons[0]) & 0x0000FF00) >> 8)*alpha; 
		b= (((CARD32)icons[0]) & 0x000000FF)*alpha; 
				
		/* Composite alpha channel: fgcolor*alpha + (bgcolor*(1-alpha)*/
		buffer_img[img_idx]= 0; 
		buffer_img[img_idx+1]= r+(rb*(1-alpha)); 
		buffer_img[img_idx+2]= g+(gb*(1-alpha)); 
		buffer_img[img_idx+3]= b+(bb*(1-alpha)); 
		img_idx+=4;
		icons+=1; /* Next scanline */
	}	

	image=XCreateImage(display, VISUAL, DEPTH, ZPixmap, 0, (char *)buffer_img, icon->w, icon->h, 32, 0);
	if (!image) BUG();
	image->byte_order=MSBFirst;
	image->bitmap_bit_order=MSBFirst;
	icon->pixmap=XCreatePixmap(display, ROOT, icon->w, icon->h, DEPTH);
	XPutImage(display, icon->pixmap, xgc, image, 0, 0, 0, 0, icon->w, icon->h);

	/* TODO: Free icons property (data pointer) */
	
	return icon;
}


void x_destroy_pixmap(xImage icon)
{
	if (icon->pixmap!=None) XFreePixmap(display, icon->pixmap);
	if (icon->mask!=None) XFreePixmap(display, icon->mask);

	free(icon);
}
    

/* Get data associated to an X atom */
unsigned char *x_get_property(Window window, Atom atom, Atom type, unsigned long *nitems)
{
	unsigned char *data;
	Atom rettype;
	int retformat;
	unsigned long nbytes;
	XTextProperty text;

	/* First look for known atoms, if not get the raw property data, the
	   caller is responsible to know what is asking for */
	switch(atom) {
		case XA_WM_NAME:
			if (XGetWMName(display, window, &text)) {
				data=text.value;
				*nitems=strlen((char *)text.value);
			} else {
				data=NULL;
				SAY("No WM_NAME found");		
			}	
			break;
		case XA_WM_ICON_NAME:
			if (XGetWMIconName(display, window, &text)) {
				data=text.value;
				*nitems=strlen((char *)text.value);
			} else {
				data=NULL;
				SAY("No WM_ICON_NAME found");		
			}	
			break;
		default:    
				if (XGetWindowProperty(display, window, atom, 0L, 65535L, False,
							   type, &rettype, &retformat,
							   nitems, &nbytes, &data)!=Success) return NULL;
				if (rettype==None) {
					SAY("No %s atom found", XGetAtomName(display, atom));
					data=NULL;
				} 
	}				   

	return data;			   
}


void x_set_property(Window window, Atom atom, Atom type, unsigned char *data, int nitems)
{
	SAY("Setting X property %s",XGetAtomName(display, atom));

	XChangeProperty(display, window, atom, type, 32, PropModeReplace, data, nitems);
}


void x_set_property_utf8(Window window, Atom atom, unsigned char *data)
{
	SAY("Setting X utf8 property %s",XGetAtomName(display, atom));

	XChangeProperty(display, window, atom, atoms.utf8_string, 8, PropModeReplace, data, strlen((char *)data));
}


/* QT names the event handlers 'slots'.  We use sloot, because we are the Puto Amo.  */
void x_addsloot (Window window, Handler handler, void *widget) 
{
	Sloot *sloot=NULL;

	ASSERT(window!=None);
	ASSERT(handler);
	ASSERT(sloots);

	/* Create the new Sloot */
	sloot=malloc(sizeof(Sloot));
	if (!sloot) die("Not enough memory");

	sloot->window=window;
	sloot->handle=handler;
	sloot->widget=widget;

	/* We can safely add duplicates :(( */
	chain_seek(sloots, chain_size(sloots));
	chain_put(sloots, sloot);
}



/* Remember: we are the Puto Amo. */
void x_delsloot (Window window, Handler handler, void *widget) 
{
	Sloot *badboy;

	ASSERT(window!=None);
	ASSERT(handler);
	ASSERT(sloots);

	chain_home(sloots);

	while ( (badboy=chain_next(sloots))!=NULL ) {
		if ((badboy->window==window) && (badboy->handle==handler) && (badboy->widget=widget)) {
			chain_del(sloots);
			free(badboy);
			return;
		}
	}
}


/* Non fatal error. We will use it for dealing with a race condition:
 * sometimes a window is destroyed while handling some event, and we
 * don't notice it, so it generates BadWindow and the like...  */
static int _errorhandler (Display *dpy, XErrorEvent *event) 
{
	char buffer[80];

	/* Assert the values */
	ASSERT(dpy);
	ASSERT(dpy==display);
	ASSERT(dpy==event->display);

	dpy=NULL;

	XGetErrorText(display,event->error_code,buffer,sizeof(buffer));

	/* Always emit this info. Helps debugging. */
	SAY("Got non fatal error on XID %#lx:\n"\
		"\tSerial number: %lu\n"\
		"\tError message: %s [%u]\n"\
		"\tRequest code: %u [%u]\n",
		event->resourceid,
		event->serial,  
		buffer, event->error_code,
		event->request_code, event->minor_code);

	/* Trying to debug an X app using the error handlers is just
	 * like smashing your own dick with a really big hammer: it may look
	 * interesting, but after a few tries you notice that is not a very
	 * good idea...  */

	return 0;
}




/* Fatal (lost connections and the like) errors */
static int _ioerrorhandler (Display *dpy) 
{
	ASSERT(dpy==display);

	/* For the sake of saying something... */
	die("Fatal IO error on display %s, goodbye :))", DisplayString(dpy));

	return 0;  /* Make gcc happy */
}


void x_main_loop(void)
{
    XEvent event;
	fd_set rfds;
	
	while(1) {
		/* Here we do our own stuff */
#ifdef ENABLE_STARTUP
		if (startup_timeout) {
			startup_timeout=false;
			XDefineCursor(display, ROOT, CURSOR);
		}
#endif

		/* Dispatch X events*/
		while (XPending(display)>0) {
			XNextEvent(display,&event);	    

			x_dispatch(&event);
		}

		/* Go to bed ;). poll is better but select is more portable. Maybe some
		   time in the future a poll call will be used... */
		FD_ZERO(&rfds); FD_SET(displayfd, &rfds);
		select(displayfd+1, &rfds, NULL, NULL, NULL);
	}

	/* ...and beyond the ligth there is only darkness... */
}


/* Dispatch one event. */
void x_dispatch (XEvent *xevent) 
{
    Window window=None;
    Sloot *sloot=NULL;
    size_t oldpos;

    /* ...determine destination... */
    window=xevent->xany.window;

    if (window==None) return;

#ifdef ENABLE_STARTUP
	sn_display_process_event (sn_display, xevent);
#endif
	
    /* Before messing with the sloots chain, save the cursor offset. This
       way we can restore it later, and allow nested x_dispatch calls */
    oldpos=chain_offset(sloots);

    /* ...and run the aproppriate handler, if any. */
    chain_home(sloots);
    while ( (sloot=chain_next(sloots))!=NULL ) {
        if (sloot->window==window) {
	    sloot->handle(xevent, sloot->widget);
	}
    }

    chain_seek(sloots, oldpos);
    
}


#ifdef ENABLE_STARTUP
static void snevent_handler (SnMonitorEvent *event, void *user_data)
{
	SnMonitorContext *context;
	SnStartupSequence *sequence;
	//const char *s;

	context = sn_monitor_event_get_context (event);
	sequence = sn_monitor_event_get_startup_sequence (event);

	switch (sn_monitor_event_get_type (event)) {
		case SN_MONITOR_EVENT_INITIATED:
		case SN_MONITOR_EVENT_CHANGED:
	
			XDefineCursor(display, ROOT, clockcursor);
			if (sn_monitor_event_get_type (event) == SN_MONITOR_EVENT_INITIATED) {
				SAY("Initiated sequence %s\n", sn_startup_sequence_get_id (sequence));
			} else {
				SAY("Changed sequence %s\n", sn_startup_sequence_get_id (sequence));
			}

#if 0
			s = sn_startup_sequence_get_id (sequence);
			SAY(" id %s", s ? s : "(unset)");
	
			s = sn_startup_sequence_get_name (sequence);
			SAY(" name %s", s ? s : "(unset)");

			s = sn_startup_sequence_get_description (sequence);
			SAY(" description %s", s ? s : "(unset)");

			SAY(" workspace %d", sn_startup_sequence_get_workspace (sequence));

			s = sn_startup_sequence_get_binary_name (sequence);
			SAY(" binary name %s", s ? s : "(unset)");
			s = sn_startup_sequence_get_icon_name (sequence);
 			s = sn_startup_sequence_get_wmclass (sequence);
		  	SAY(" wm class %s", s ? s : "(unset)");
#endif

			seq_ref_count++;
			break;

		case SN_MONITOR_EVENT_COMPLETED:
			seq_ref_count--;
			if (!seq_ref_count){
				alarm(0); /* Reset timeout */
				XDefineCursor(display, ROOT, CURSOR);
				SAY ("Completed sequence %s", sn_startup_sequence_get_id (sequence));
			}
			break;

		case SN_MONITOR_EVENT_CANCELED:
			SAY ("Canceled sequence %s", sn_startup_sequence_get_id (sequence));
			break;
	}
}
#endif

